"""
This script allows to take 4 positions and to create a workspace from them
"""

# Imports
from pyniryo import NiryoRobot

import sys

if sys.version_info[0] == 2:
    input_func = raw_input
else:
    input_func = input


robot_ip_address = "10.10.10.10"

# Connecting to robot
niyro_robot = NiryoRobot(robot_ip_address)

niyro_robot.calibrate_auto()
niyro_robot.set_learning_mode(True)

# Asking user to type the new workspace's name
ws_name = input_func("Enter name of new workspace. Name: ")

# Initializing useful variables
points = []
id_point = 1

while id_point < 5:  # Iterating over 4 markers
    input_func("Press enter when on point".format(id_point + 1))
    # Getting pose
    pose = niyro_robot.get_pose()
    points.append(pose)
    id_point += 1

# Creating workspace
niyro_robot.save_workspace_from_robot_poses(ws_name, *points)
# Leaving
niyro_robot.close_connection()



"""
Little Script which shows how to read a position with PyNiryo
"""
# !/usr/bin/env python

from pyniryo import NiryoRobot
import sys

if sys.version_info[0] == 2:
    input_func = raw_input
else:
    input_func = input


# Connecting to robot
niryo_robot = NiryoRobot(ip_address="10.10.10.10")

niryo_robot.calibrate_auto()

while "User to not quit":
    # Going to learning mode
    niryo_robot.set_learning_mode(True)

    input_func("Press enter to switch to turn learning mode off")
    niryo_robot.set_learning_mode(False)

    # Reading pose and displaying it
    pose = niryo_robot.get_pose()
    print(pose)
    if input_func("Press enter to get new pose / q to leave ") == 'q':
        break
niryo_robot.set_learning_mode(True)

niryo_robot.close_connection()



#!/usr/bin/env python
import math


def get_angle_2_points(x1, y1, x2, y2):
    """
    angle of the line starting at (x1,y1) and finishing at (x2,y2), according to repere trigonometrique
    :param x1: x position of the first point
    :param y1: y position of the first point
    :param x2: x position of the second point
    :param y2: y position of the second point
    :return: angle in radians
    """
    dy = float(y2) - float(y1)
    dx = float(x2) - float(x1)
    if dx == 0:
        # return 90 if dy > 0 else -90
        return math.pi / 2 if dy > 0 else -math.pi / 2

    else:
        return math.atan2(dy, dx)


def euclidean_dist(x1, y1, x2, y2):
    """
    Return euclidean distance between 2 points
    :param x1: x position of the first point
    :param y1: y position of the first point
    :param x2: x position of the second point
    :param y2: y position of the second point
    :return: distance in the same metrics as the points
    """
    return math.sqrt((float(x1) - float(x2)) ** 2 + (float(y1) - float(y2)) ** 2)


def euclidean_dist_2_pts(p1, p2):
    """
    Return euclidean distance between 2 points
    :param p1: tuple(X,Y) of the first point's coordinates
    :param p2: tuple(X,Y) of the second point's coordinates
    :return: distance in the same metrics as the points
    """
    x1, y1 = p1
    x2, y2 = p2
    return math.sqrt((float(x1) - float(x2)) ** 2 + (float(y1) - float(y2)) ** 2)



"""
Exception classes for TCP
"""


class NiryoRobotException(Exception):
    pass


class ClientNotConnectedException(Exception):
    def __init__(self, msg=None):
        if msg is None:
            msg = "You're not connected to  the robot."
        super(Exception, self).__init__(msg)


class InvalidAnswerException(Exception):
    def __init__(self, answer):
        super(Exception, self).__init__("An invalid answer has been received. Format expected: JSON.\n"
                                        + "A problem occurred with: '" + answer + "'")


class TcpCommandException(Exception):
    pass


class HostNotReachableException(Exception):
    def __init__(self):
        super(Exception, self).__init__("Unable to communicate with robot server, please verify your network.")



from enum import Enum, unique
import cv2

# -- Text Display
RED = (0, 0, 255)
PURPLE = (255, 0, 255)
ORANGE = (0, 127, 255)
GREEN = (0, 255, 0)
BLUE = (255, 0, 0)
WHITE = (255, 255, 255)
BLACK = (0, 0, 0)
font = cv2.FONT_HERSHEY_SIMPLEX
font_scale_big = 2.0
font_scale_normal = 0.7
thickness_big = 3
thickness_small = 2


# Values for HSV threshold
@unique
class ColorHSV(Enum):
    """
    MIN HSV, MAX HSV, Invert Hue (bool)
    """
    BLUE = [90, 50, 85], [125, 255, 255], False
    RED = [15, 80, 75], [170, 255, 255], True
    GREEN = [40, 60, 75], [85, 255, 255], False
    ANY = [0, 50, 100], [179, 255, 255], False


# Values use for HS'V thresholding
@unique
class ColorHSVPrime(Enum):
    """
    MIN HSV, MAX HSV, Invert Hue (bool)
    """
    BLUE = [90, 70, 100], [115, 255, 255], False
    RED = [15, 70, 100], [170, 255, 255], True
    GREEN = [40, 70, 100], [85, 255, 255], False
    ANY = [0, 70, 140], [179, 255, 255], False


# Different object types handle by the API. The value corresponds to the object's number of sides
@unique
class ObjectType(Enum):
    SQUARE = 4
    TRIANGLE = 3
    CIRCLE = -1
    ANY = 0


@unique
class MorphoType(Enum):
    ERODE = cv2.MORPH_ERODE
    DILATE = cv2.MORPH_DILATE
    OPEN = cv2.MORPH_OPEN
    CLOSE = cv2.MORPH_CLOSE


@unique
class KernelType(Enum):
    RECT = cv2.MORPH_RECT
    ELLIPSE = cv2.MORPH_ELLIPSE
    CROSS = cv2.MORPH_CROSS



from pyniryo import NiryoRobot  # Niryo's Python Package
import time
from pynput.mouse import Button, Listener, Controller  # Package to get mouse inputs

simulation_mode = True

robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simulation = "127.0.0.1"

# Set robot address
robot_ip_address = robot_ip_address_simulation if simulation_mode else robot_ip_address_rpi


class MyMouseListener:

    def __init__(self):
        self.clicked = False

        self.last_x_send = None
        self.actual_x = None

        self.last_y_send = None
        self.actual_y = None

        self.last_z_send = 0
        self.actual_z = 0

        self.mouse = Controller()

        self.listener = Listener(
            on_move=self.on_move,
            on_click=self.on_click,
            on_scroll=self.on_scroll).start()

    def __del__(self):
        self.listener.stop()

    def on_move(self, x, y):
        if self.last_x_send is None:
            self.last_x_send = x
            self.last_y_send = y
        self.actual_x = x
        self.actual_y = y

    def on_click(self, _x, _y, button, pressed):
        if not pressed:
            # self.mouse.release(button)
            if button == Button.right:
                self.clicked = True
            elif button == Button.left:
                self.actual_x = 960
                self.actual_y = 1500
                self.mouse.position = (self.actual_x, self.actual_y)
                self.actual_z = 0
                self.reset()

    def on_scroll(self, _x, _y, _dx, dy):
        self.actual_z += dy

    def reset(self):
        self.last_x_send = self.actual_x
        self.last_y_send = self.actual_y
        self.last_z_send = self.actual_z

    def get_x_diff(self):
        if self.last_x_send is None:
            return 0
        return self.actual_x - self.last_x_send

    def get_y_diff(self):
        if self.last_y_send is None:
            return 0
        return self.actual_y - self.last_y_send

    def get_z_diff(self):
        return self.actual_z - self.last_z_send

    def __str__(self):
        return ''


def process(niyro_robot):
    mouse_listener = MyMouseListener()
    niyro_robot.move_joints(0.0, 0.0, 0.0, 0.0, -1.57, 0.0)
    niyro_robot.set_jog_control(True)
    while not mouse_listener.clicked:
        init = time.time()
        mouse_dx = mouse_listener.get_x_diff()
        mouse_dy = mouse_listener.get_y_diff()
        mouse_dz = mouse_listener.get_z_diff()
        mouse_listener.reset()
        robot_dx = - float(mouse_dy) * 1e-4
        robot_dy = - float(mouse_dx) * 1e-4
        robot_dz = float(mouse_dz) * 5e-3
        niyro_robot.jog_pose(robot_dx, robot_dy, robot_dz, 0.0, 0.0, 0.0)

        time.sleep(max(0., 0.10 - (time.time() - init)))

    niyro_robot.set_jog_control(False)
    niyro_robot.set_learning_mode(True)


if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Calibrate robot if robot needs calibration
    robot.calibrate_auto()
    # Launching main process
    process(robot)
    # Releasing connection
    robot.close_connection()



# Imports
from pyniryo import *

simulation_mode = True

# Set robot address & workspace name
if simulation_mode:
    robot_ip_address, workspace_name = "127.0.0.1", "gazebo_1"
else:
    robot_ip_address, workspace_name = "10.10.10.10", "workspace_1"

# -- Should Change these variables
# The pose from where the image processing happens
observation_pose = PoseObject(
    x=0.18, y=0.0, z=0.35,
    roll=0.0, pitch=1.57, yaw=-0.2,
)

# First conditioning pose
place_pose_simulation = PoseObject(
    x=0.31, y=-0.12, z=0.09,
    roll=0, pitch=1.57, yaw=0.0
)
place_pose_reality = PoseObject(
    x=0.30, y=-0.105, z=0.1,
    roll=0, pitch=1.57, yaw=0.0
)
offset_y_between_place = -0.055

if simulation_mode:
    place_pose = place_pose_simulation
else:
    place_pose = place_pose_reality


# -- MAIN PROGRAM

def vision_pick_n_place(niyro_robot):
    # Loop
    try_without_success = 0
    catch_count = 0
    while try_without_success < 3:
        # Moving to observation pose
        niyro_robot.move_pose(observation_pose)
        # Trying to pick target using camera
        ret = niyro_robot.vision_pick(workspace_name,
                                      height_offset=0.0,
                                      shape=ObjectShape.ANY,
                                      color=ObjectColor.ANY)
        obj_found, shape_ret, color_ret = ret
        if not obj_found:
            try_without_success += 1
            continue

        # Everything is good, so we going to place the object
        new_place_pose = place_pose.copy_with_offsets(x_offset=offset_y_between_place * catch_count)
        niyro_robot.place_from_pose(new_place_pose)
        catch_count += 1


if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Calibrate robot if robot needs calibration
    robot.calibrate_auto()
    # Update tool
    robot.update_tool()
    # Launching main process
    vision_pick_n_place(robot)
    # Releasing connection
    robot.close_connection()



# - Imports
from __future__ import print_function

# Python libraries
import time
import socket
import numpy as np
import sys

# Communication imports
from .enums_communication import *
from .communication_functions import dict_to_packet, receive_dict, receive_dict_w_payload

from .exceptions import *
from .objects import PoseObject, HardwareStatusObject, DigitalPinObject, AnalogPinObject


class NiryoRobot(object):
    def __init__(self, ip_address=None):
        self.__ip_address = None
        self.__port = TCP_PORT
        self.__client_socket = None

        self.__timeout = TCP_TIMEOUT

        self.__is_running = True
        self.__is_connected = False

        # If user give IP Address, try to connect directly
        if ip_address is not None:
            self.connect(ip_address)

    def __del__(self):
        self.close_connection()

    def __str__(self):
        if self.__is_connected:
            msg = "\nConnected to server ({}) on port: {}\n".format(self.__ip_address, self.__port)
        else:
            msg = "Not Connected"

        return msg

    def __repr__(self):
        return self.__str__()

    # -- Connection
    def connect(self, ip_address):
        """
        Connect to the TCP Server

        :param ip_address: IP Address
        :type ip_address: str
        :rtype: None
        """
        self.__client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__client_socket.settimeout(self.__timeout)
        try:
            self.__client_socket.connect((ip_address, self.__port))
        except (socket.timeout, socket.error) as e:
            self.__client_socket = None
            raise ClientNotConnectedException("Unable to connect to the robot : {}".format(e))

        self.__is_connected = True
        self.__client_socket.settimeout(None)
        self.__ip_address = ip_address
        print("\nConnected to server ({}) on port: {}\n".format(ip_address, self.__port))

    def __close_socket(self):
        if self.__client_socket is not None and self.__is_connected is True:
            try:
                self.__client_socket.shutdown(socket.SHUT_RDWR)
                self.__client_socket.close()
            except socket.error:
                pass
            self.__is_connected = False
            print("\nDisconnected from robot\n")

    def close_connection(self):
        """
        Close connection with robot

        :rtype: None
        """
        self.__is_running = False
        self.__close_socket()
        self.__client_socket = None

    # -- SEND & RECEIVE

    # - Sending phase
    def __send_command(self, command_type, *parameter_list):
        if self.__is_connected is False:
            raise ClientNotConnectedException()
        if self.__client_socket is not None:
            try:
                ret_dict = self.__build_dict(command_type, *parameter_list)
                packet = dict_to_packet(ret_dict)
                self.__client_socket.send(packet)
            except socket.error as e:
                print(e, file=sys.stderr)
                raise HostNotReachableException()

    @staticmethod
    def __build_dict(command_type, *parameter_list):
        command = command_type.name
        new_param_list = []
        for parameter in parameter_list:
            if parameter is None:
                new_param_list.append("None")
            elif isinstance(parameter, Enum):
                new_param_list.append(parameter.name)
            elif isinstance(parameter, bool):
                new_param_list.append(str(parameter).upper())
            else:
                new_param_list.append(parameter)

        return {"command": command, "param_list": new_param_list}

    # - Receiving

    def __receive_answer(self, with_payload):
        try:
            if not with_payload:
                received_dict = receive_dict(sckt=self.__client_socket)
                payload = None
            else:
                received_dict, payload = receive_dict_w_payload(sckt=self.__client_socket)
        except socket.error as e:
            print(e, file=sys.stderr)
            raise HostNotReachableException()
        if not received_dict:
            raise HostNotReachableException()
        answer_status = received_dict["status"]
        if answer_status not in ["OK", "KO"]:
            raise InvalidAnswerException(answer_status)
        if received_dict["status"] != "OK":
            raise NiryoRobotException("Command KO : {}".format(received_dict["message"]))
        list_ret_param = received_dict["list_ret_param"]
        if len(list_ret_param) == 1:
            list_ret_param = list_ret_param[0]
        if not with_payload:
            return list_ret_param
        else:
            return list_ret_param, payload

    # - Wrapping functions
    def __send_n_receive(self, command_type, *parameter_list, **kwargs):
        with_payload = kwargs.get("with_payload", False)

        self.__send_command(command_type, *parameter_list)
        return self.__receive_answer(with_payload=with_payload)

    # Parameters checker
    def __check_enum_belonging(self, value, enum_):
        """
        Check if a value belong to an enum
        """
        if value not in enum_:
            self.__raise_exception_expected_choice([v for v in enum_], value)

    def __check_list_belonging(self, value, list_):
        """
        Check if a value belong to a list
        """
        if value not in list_:
            self.__raise_exception_expected_choice(list_, value)

    def __check_range_belonging(self, value, range_min, range_max):
        """
        Check if a value belong to a range
        """
        if not range_min <= value <= range_max:
            self.__raise_exception_expected_range(range_min, range_max, value)

    def __check_dict_belonging(self, value, dict_):
        """
        Check if a value belong to a dictionary
        """
        if value not in dict_.keys():
            self.__raise_exception_expected_choice(dict_.keys(), value)

    def __check_type(self, value, type_):
        if type(value) is not type_:
            self.__raise_exception_expected_type(type_.__name__, value)

    def __check_instance(self, value, type_):
        if not isinstance(value, type_):
            self.__raise_exception_expected_type(type_.__name__, value)

    def __check_list_type(self, values_list, type_):
        for value in values_list:
            self.__check_type(value, type_)

    def __map_list(self, list_, type_):
        """
        Try to map a list to another type (Very useful for list like joints
        which are acquired as string)
        """
        try:
            map_list = list(map(type_, list_))
            return map_list
        except ValueError:
            self.__raise_exception_expected_type(type_.__name__, list_)

    def __transform_to_type(self, value, type_):
        """
        Try to change value type to another
        """
        try:
            value = type_(value)
            return value
        except ValueError:
            self.__raise_exception_expected_type(type_.__name__, value)

    # Error Handlers
    def __raise_exception_expected_choice(self, expected_choice, given):
        raise TcpCommandException("Expected one of the following: {}.\nGiven: {}".format(expected_choice, given))

    def __raise_exception_expected_type(self, expected_type, given):
        raise TcpCommandException("Expected type: {}.\nGiven: {}".format(expected_type, given))

    def __raise_exception_expected_range(self, range_min, range_max, given):
        raise TcpCommandException(
            "Expected the following condition: {} <= value <= {}\nGiven: {}".format(range_min, range_max, given))

    def __raise_exception(self, message):
        raise TcpCommandException("Exception message : {}".format(message))

    # -- Useful functions
    def __args_pose_to_list(self, *args):
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, PoseObject):
                return arg.to_list()
            else:
                pose_list = arg
        else:
            pose_list = args

        pose_list_float = self.__map_list(pose_list, float)
        if len(pose_list_float) != 6:
            self.__raise_exception("A pose should contain 6 elements (x, y, z, roll, pitch, yaw)")
        return pose_list_float

    def __args_joints_to_list(self, *args):
        """
        Convert args into a list
        Either if args = (1.1,5.6,-6.7) or args = ([1.1,5.6,-6.7],) , the
        function will return (1.1,5.6,-6.7)

        :param args: Union[list, tuple]
        :return: list of float
        """
        if len(args) == 1:
            args = args[0]

        joints = self.__map_list(args, float)
        if len(joints) != 6:
            self.__raise_exception("The robot has 6 joints")

        return joints

    # --- Public functions --- #

    # - Main purpose

    def calibrate(self, calibrate_mode):
        """
        Calibrate (manually or automatically) motors. Automatic calibration will do nothing
        if motors are already calibrated

        :param calibrate_mode: Auto or Manual
        :type calibrate_mode: CalibrateMode
        :rtype: None
        """
        self.__check_enum_belonging(calibrate_mode, CalibrateMode)
        self.__send_n_receive(Command.CALIBRATE, calibrate_mode)

    def calibrate_auto(self):
        """
        Start a automatic motors calibration if motors are not calibrated yet

        :rtype: None
        """
        self.calibrate(CalibrateMode.AUTO)

    def need_calibration(self):
        """
        Return a bool indicating whereas the robot motors need to be calibrate

        :rtype: bool
        """
        hardware_status = self.get_hardware_status()
        return hardware_status.calibration_needed

    @property
    def learning_mode(self):
        """
        Property
        Get learning mode state

        Example: ::

            if robot.learning_mode:
                print("Torque off")
            else:
                print("Torque on")

        :return: ``True`` if learning mode is on
        :rtype: bool
        """
        return self.get_learning_mode()

    def get_learning_mode(self):
        """
        Get learning mode state

        :return: ``True`` if learning mode is on
        :rtype: bool
        """
        return eval(self.__send_n_receive(Command.GET_LEARNING_MODE))

    @learning_mode.setter
    def learning_mode(self, value):
        self.set_learning_mode(value)

    def set_learning_mode(self, enabled):
        """
        Set learning mode if param is ``True``, else turn it off

        :param enabled: ``True`` or ``False``
        :type enabled: bool
        :rtype: None
        """
        self.__check_type(enabled, bool)
        self.__send_n_receive(Command.SET_LEARNING_MODE, enabled)

    def set_arm_max_velocity(self, percentage_speed):
        """
        Limit arm max velocity to a percentage of its maximum velocity

        :param percentage_speed: Should be between 1 & 100
        :type percentage_speed: int
        :rtype: None
        """
        self.__check_range_belonging(percentage_speed, 1, 100)
        self.__send_n_receive(Command.SET_ARM_MAX_VELOCITY, percentage_speed)

    def set_jog_control(self, enabled):
        """
        Set jog control mode if param is True, else turn it off

        :param enabled: ``True`` or ``False``
        :type enabled: bool
        :rtype: None
        """
        self.__check_type(enabled, bool)
        self.__send_n_receive(Command.SET_JOG_CONTROL, enabled)

    @staticmethod
    def wait(duration):
        """
        Wait for a certain time

        :param duration: duration in seconds
        :type duration: float
        :rtype: None
        """
        time.sleep(duration)

    # - Joints/Pose

    @property
    def joints(self):
        return self.get_joints()

    def get_joints(self):
        """
        Get joints value in radians
        You can also use a getter ::

            joints = robot.get_joints()
            joints = robot.joints

        :return: List of joints value
        :rtype: list[float]
        """
        return self.__send_n_receive(Command.GET_JOINTS)

    @property
    def pose(self):
        return self.get_pose()

    def get_pose(self):
        """
        Get end effector link pose as [x, y, z, roll, pitch, yaw].
        x, y & z are expressed in meters / roll, pitch & yaw are expressed in radians
        You can also use a getter ::

            pose = robot.get_pose()
            pose = robot.pose

        :rtype: PoseObject
        """
        data = self.__send_n_receive(Command.GET_POSE)
        pose_array = self.__map_list(data, float)
        pose_object = PoseObject(*pose_array)
        return pose_object

    def get_pose_quat(self):
        """
        Get end effector link pose in Quaternion coordinates

        :return: Position and quaternion coordinates concatenated in a list : [x, y, z, qx, qy, qz, qw]
        :rtype: list[float]
        """
        data = self.__send_n_receive(Command.GET_POSE_QUAT)
        pose_array = self.__map_list(data, float)
        return pose_array

    @joints.setter
    def joints(self, *args):
        self.move_joints(*args)

    def move_joints(self, *args):
        """
        Move robot joints. Joints are expressed in radians.

        All lines of the next example realize the same operation: ::

            robot.joints = [0.2, 0.1, 0.3, 0.0, 0.5, 0.0]
            robot.move_joints([0.2, 0.1, 0.3, 0.0, 0.5, 0.0])
            robot.move_joints(0.2, 0.1, 0.3, 0.0, 0.5, 0.0)

        :param args: either 6 args (1 for each joints) or a list of 6 joints
        :type args: Union[list[float], tuple[float]]
        :rtype: None
        """
        joints = self.__args_joints_to_list(*args)
        self.__send_n_receive(Command.MOVE_JOINTS, *joints)

    @pose.setter
    def pose(self, *args):
        self.move_pose(*args)

    def move_pose(self, *args):
        """
        Move robot end effector pose to a (x, y, z, roll, pitch, yaw) pose.
        x, y & z are expressed in meters / roll, pitch & yaw are expressed in radians

        All lines of the next example realize the same operation: ::

            robot.pose = [0.2, 0.1, 0.3, 0.0, 0.5, 0.0]
            robot.move_pose([0.2, 0.1, 0.3, 0.0, 0.5, 0.0])
            robot.move_pose(0.2, 0.1, 0.3, 0.0, 0.5, 0.0)
            robot.move_pose(PoseObject(0.2, 0.1, 0.3, 0.0, 0.5, 0.0))

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a ``PoseObject``
        :type args: Union[tuple[float], list[float], PoseObject]

        :rtype: None
        """
        pose_list = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.MOVE_POSE, *pose_list)

    def move_linear_pose(self, *args):
        """
        Move robot end effector pose to a (x, y, z, roll, pitch, yaw) pose with a linear trajectory

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[tuple[float], list[float], PoseObject]
        :rtype: None
        """
        pose_list = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.MOVE_LINEAR_POSE, *pose_list)

    def shift_pose(self, axis, shift_value):
        """
        Shift robot end effector pose along one axis

        :param axis: Axis along which the robot is shifted
        :type axis: RobotAxis
        :param shift_value: In meter for X/Y/Z and radians for roll/pitch/yaw
        :type shift_value: float
        :rtype: None
        """
        self.__check_enum_belonging(axis, RobotAxis)
        shift_value = self.__transform_to_type(shift_value, float)
        self.__send_n_receive(Command.SHIFT_POSE, axis, shift_value)

    def shift_linear_pose(self, axis, shift_value):
        """
        Shift robot end effector pose along one axis, with a linear trajectory

        :param axis: Axis along which the robot is shifted
        :type axis: RobotAxis
        :param shift_value: In meter for X/Y/Z and radians for roll/pitch/yaw
        :type shift_value: float
        :rtype: None
        """
        self.__check_enum_belonging(axis, RobotAxis)
        shift_value = self.__transform_to_type(shift_value, float)
        self.__send_n_receive(Command.SHIFT_LINEAR_POSE, axis, shift_value)

    def jog_joints(self, *args):
        """
        Jog robot joints'.
        Jog corresponds to a shift without motion planning.
        Values are expressed in radians.

        :param args: either 6 args (1 for each joints) or a list of 6 joints offset
        :type args: Union[list[float], tuple[float]]
        :rtype: None
        """
        joints_offset = self.__args_joints_to_list(*args)
        self.__send_n_receive(Command.JOG_JOINTS, *joints_offset)

    def jog_pose(self, *args):
        """
        Jog robot end effector pose
        Jog corresponds to a shift without motion planning
        Arguments are [dx, dy, dz, d_roll, d_pitch, d_yaw]
        dx, dy & dz are expressed in meters / d_roll, d_pitch & d_yaw are expressed in radians

        :param args: either 6 args (1 for each coordinates) or a list of 6 offset
        :type args: Union[list[float], tuple[float]]
        :rtype: None
        """
        pose_offset = self.__args_joints_to_list(*args)
        self.__send_n_receive(Command.JOG_POSE, *pose_offset)

    def move_to_home_pose(self):
        """
        Move to a position where the forearm lays on shoulder

        :rtype: None
        """
        self.move_joints(0.0, 0.3, -1.3, 0.0, 0.0, 0.0)

    def go_to_sleep(self):
        """
        Go to home pose and activate learning mode

        :rtype: None
        """
        self.move_to_home_pose()
        self.set_learning_mode(True)

    def forward_kinematics(self, *args):
        """
        Compute forward kinematics of a given joints configuration and give the
        associated spatial pose

        :param args: either 6 args (1 for each joints) or a list of 6 joints
        :type args: Union[list[float], tuple[float]]
        :rtype: PoseObject
        """
        joints = self.__args_joints_to_list(*args)
        data = self.__send_n_receive(Command.FORWARD_KINEMATICS, *joints)

        pose_array = self.__map_list(data, float)
        pose_object = PoseObject(*pose_array)
        return pose_object

    def inverse_kinematics(self, *args):
        """
        Compute inverse kinematics

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a ``PoseObject``
        :type args: Union[tuple[float], list[float], PoseObject]

        :return: List of joints value
        :rtype: list[float]
        """
        pose_list = self.__args_pose_to_list(*args)

        return self.__send_n_receive(Command.INVERSE_KINEMATICS, *pose_list)

    # - Saved Pose

    def get_pose_saved(self, pose_name):
        """
        Get pose saved in from Ned's memory
        
        :param pose_name: Pose name in robot's memory
        :type pose_name: str 
        :return: Pose associated to pose_name
        :rtype: PoseObject
        """
        self.__check_type(pose_name, str)

        pose_list = self.__send_n_receive(Command.GET_POSE_SAVED, pose_name)
        return PoseObject(*pose_list)

    def save_pose(self, pose_name, *args):
        """
        Save pose in robot's memory
        
        :type pose_name: str
        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None 
        """
        self.__check_type(pose_name, str)
        pose_list = self.__args_pose_to_list(*args)

        self.__send_n_receive(Command.SAVE_POSE, pose_name, *pose_list)

    def delete_pose(self, pose_name):
        """
        Delete pose from robot's memory

        :type pose_name: str
        :rtype: None
        """
        self.__check_type(pose_name, str)
        self.__send_n_receive(Command.DELETE_POSE, pose_name)

    def get_saved_pose_list(self):
        """
        Get list of poses' name saved in robot memory

        :rtype: list[str]
        """
        return self.__send_n_receive(Command.GET_SAVED_POSE_LIST)

    # - Pick/Place

    def pick_from_pose(self, *args):
        """
        Execute a picking from a pose.

        A picking is described as : \n
        | * going over the object
        | * going down until height = z
        | * grasping with tool
        | * going back over the object

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None
        """
        pose = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.PICK_FROM_POSE, *pose)

    def place_from_pose(self, *args):
        """
        Execute a placing from a position.
        
        A placing is described as : \n
        | * going over the place
        | * going down until height = z
        | * releasing the object with tool
        | * going back over the place

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None
        """
        pose = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.PLACE_FROM_POSE, *pose)

    def pick_and_place(self, pick_pose, place_pos, dist_smoothing=0.0):
        """
        Execute a pick then a place

        :param pick_pose: Pick Pose : [x, y, z, roll, pitch, yaw] or PoseObject
        :type pick_pose: Union[list[float], PoseObject]
        :param place_pos: Place Pose : [x, y, z, roll, pitch, yaw] or PoseObject
        :type place_pos: Union[list[float], PoseObject]
        :param dist_smoothing: Distance from waypoints before smoothing trajectory
        :type dist_smoothing: float
        :rtype: None
        """
        pick_pose_list = self.__args_pose_to_list(pick_pose)
        place_pos_list = self.__args_pose_to_list(place_pos)

        self.__send_n_receive(Command.PICK_AND_PLACE, pick_pose_list, place_pos_list, dist_smoothing)

    # - Trajectories
    def get_trajectory_saved(self, trajectory_name):
        """
        Get trajectory saved in Ned's memory

        :type trajectory_name: str
        :return: Trajectory
        :rtype: list[list[float]]
        """
        self.__check_type(trajectory_name, str)
        return self.__send_n_receive(Command.GET_TRAJECTORY_SAVED, trajectory_name)

    def execute_trajectory_saved(self, trajectory_name):
        """
        Execute trajectory from Ned's memory

        :type trajectory_name: str
        :rtype: None
        """
        self.__check_type(trajectory_name, str)
        self.__send_n_receive(Command.EXECUTE_TRAJECTORY_SAVED, trajectory_name)

    def execute_trajectory_from_poses(self, list_poses, dist_smoothing=0.0):
        """
        Execute trajectory from list of poses

        :param list_poses: List of [x,y,z,qx,qy,qz,qw] or list of [x,y,z,roll,pitch,yaw]
        :type list_poses: list[list[float]]
        :param dist_smoothing: Distance from waypoints before smoothing trajectory
        :type dist_smoothing: float
        :rtype: None
        """
        for i, pose in enumerate(list_poses):
            if len(pose) != 7 and len(pose) != 6:
                self.__raise_exception(
                    "7 parameters expected in a pose [x,y,z,qx,qy,qz,qw], or 6 in a pose [x,y,z,roll,pitch,yaw], "
                    "{} parameters given".format(len(pose)))
            list_poses[i] = self.__map_list(pose, float)

        self.__send_n_receive(Command.EXECUTE_TRAJECTORY_FROM_POSES, list_poses, dist_smoothing)

    def execute_trajectory_from_poses_and_joints(self, list_pose_joints, list_type=None, dist_smoothing=0.0):
        """
        Execute trajectory from list of poses and joints

        Example: ::

            robot.execute_trajectory_from_poses_and_joints(
               list_pose_joints = [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.25, 0.0, 0.0, 0.0, 0.0, 0.0]],
               list_type = ['joint', 'pose'],
               dist_smoothing = 0.01
            )

        :param list_pose_joints: List of [x,y,z,qx,qy,qz,qw] or list of [x,y,z,roll,pitch,yaw]
                                        or a list of [j1,j2,j3,j4,j5,j6]
        :type list_pose_joints: list[list[float]]
        :param list_type: List of string 'pose' or 'joint', or ['pose'] (if poses only) or ['joint'] (if joints only).
                        If None, it is assumed there are only poses in the list.
        :type list_type: list[string]
        :param dist_smoothing: Distance from waypoints before smoothing trajectory
        :type dist_smoothing: float
        :rtype: None
        """
        if list_type is None:
            list_type = ['pose']
        for i, pose_or_joint in enumerate(list_pose_joints):
            if len(pose_or_joint) != 7 and len(pose_or_joint) != 6:
                self.__raise_exception(
                    "7 parameters expected in a pose [x,y,z,qx,qy,qz,qw], or 6 in a pose [x,y,z,roll,pitch,yaw], "
                    "or 6 in a joint [j1,j2,j3,j4,j5,j6]"
                    "{} parameters given".format(len(pose_or_joint)))
            list_pose_joints[i] = self.__map_list(pose_or_joint, float)
        self.__send_n_receive(Command.EXECUTE_TRAJECTORY_FROM_POSES_AND_JOINTS, list_pose_joints, list_type,
                              dist_smoothing)

    def save_trajectory(self, trajectory_name, list_poses):
        """
        Save trajectory in robot memory

        :type trajectory_name: str
        :param list_poses: List of [x,y,z,qx,qy,qz,qw] or list of [x,y,z,roll,pitch,yaw]
        :type list_poses: list[list[float]]
        :rtype: None
        """
        self.__check_type(trajectory_name, str)
        for i, pose in enumerate(list_poses):
            if len(pose) != 7 and len(pose) != 6:
                self.__raise_exception(
                    "7 parameters expected in a pose [x,y,z,qx,qy,qz,qw], or 6 in a pose [x,y,z,roll,pitch,yaw], "
                    "{} parameters given".format(len(pose)))
            list_poses[i] = self.__map_list(pose, float)

        self.__send_n_receive(Command.SAVE_TRAJECTORY, trajectory_name, list_poses)

    def delete_trajectory(self, trajectory_name):
        """
        Delete trajectory from robot's memory

        :type trajectory_name: str
        :rtype: None
        """
        self.__check_type(trajectory_name, str)
        self.__send_n_receive(Command.DELETE_TRAJECTORY, trajectory_name)

    def get_saved_trajectory_list(self):
        """
        Get list of trajectories' name saved in robot memory

        :rtype: list[str]
        """
        return self.__send_n_receive(Command.GET_SAVED_TRAJECTORY_LIST)

    # -- Tools

    @property
    def tool(self):
        return self.get_current_tool_id()

    def get_current_tool_id(self):
        """
        Get equipped tool Id

        :rtype: ToolID
        """
        tool_id = self.__send_n_receive(Command.GET_CURRENT_TOOL_ID)
        return ToolID[tool_id]

    def update_tool(self):
        """
        Update equipped tool

        :rtype: None
        """
        self.__send_n_receive(Command.UPDATE_TOOL)

    def grasp_with_tool(self):
        """
        Grasp with tool
        | This action correspond to
        | - Close gripper for Grippers
        | - Pull Air for Vacuum pump
        | - Activate for Electromagnet

        :rtype: None
        """
        self.__send_n_receive(Command.GRASP_WITH_TOOL)

    def release_with_tool(self):
        """
        Release with tool
        | This action correspond to
        | - Open gripper for Grippers
        | - Push Air for Vacuum pump
        | - Deactivate for Electromagnet

        :rtype: None
        """
        self.__send_n_receive(Command.RELEASE_WITH_TOOL)

    # - Gripper
    def open_gripper(self, speed=500, max_torque_percentage=100, hold_torque_percentage=30):
        """
        Open gripper

        :param speed: Between 100 & 1000 (only for Niryo One and Ned1)
        :type speed: int
        :param max_torque_percentage: Closing torque percentage (only for Ned2)
        :type max_torque_percentage: int
        :param hold_torque_percentage: Hold torque percentage after closing (only for Ned2)
        :type hold_torque_percentage: int
        :rtype: None
        """
        speed = self.__transform_to_type(speed, int)
        max_torque_percentage = self.__transform_to_type(max_torque_percentage, int)
        hold_torque_percentage = self.__transform_to_type(hold_torque_percentage, int)

        self.__send_n_receive(Command.OPEN_GRIPPER, speed, max_torque_percentage, hold_torque_percentage)

    def close_gripper(self, speed=500, max_torque_percentage=100, hold_torque_percentage=20):
        """
        Close gripper

        :param speed: Between 100 & 1000 (only for Niryo One and Ned1)
        :type speed: int
        :param max_torque_percentage: Opening torque percentage (only for Ned2)
        :type max_torque_percentage: int
        :param hold_torque_percentage: Hold torque percentage after opening (only for Ned2)
        :type hold_torque_percentage: int
        :rtype: None
        """
        speed = self.__transform_to_type(speed, int)
        max_torque_percentage = self.__transform_to_type(max_torque_percentage, int)
        hold_torque_percentage = self.__transform_to_type(hold_torque_percentage, int)

        self.__send_n_receive(Command.CLOSE_GRIPPER, speed, max_torque_percentage, hold_torque_percentage)

    # - Vacuum
    def pull_air_vacuum_pump(self):
        """
        Pull air of vacuum pump

        :rtype: None
        """
        self.__send_n_receive(Command.PULL_AIR_VACUUM_PUMP)

    def push_air_vacuum_pump(self):
        """
        Push air of vacuum pump

        :rtype: None
        """
        self.__send_n_receive(Command.PUSH_AIR_VACUUM_PUMP)

    # - Electromagnet
    def setup_electromagnet(self, pin_id):
        """
        Setup electromagnet on pin

        :param pin_id:
        :type pin_id: PinID or str
        :rtype: None
        """
        self.__check_instance(pin_id, (PinID, str))
        pin_id_str = pin_id.value if isinstance(pin_id, PinID) else pin_id
        self.__send_n_receive(Command.SETUP_ELECTROMAGNET, pin_id_str)

    def activate_electromagnet(self, pin_id):
        """
        Activate electromagnet associated to electromagnet_id on pin_id

        :param pin_id:
        :type pin_id: PinID or str
        :rtype: None
        """
        self.__check_instance(pin_id, (PinID, str))
        pin_id_str = pin_id.value if isinstance(pin_id, PinID) else pin_id
        self.__send_n_receive(Command.ACTIVATE_ELECTROMAGNET, pin_id_str)

    def deactivate_electromagnet(self, pin_id):
        """
        Deactivate electromagnet associated to electromagnet_id on pin_id

        :param pin_id:
        :type pin_id: PinID or str
        :rtype: None
        """
        self.__check_instance(pin_id, (PinID, str))
        pin_id_str = pin_id.value if isinstance(pin_id, PinID) else pin_id
        self.__send_n_receive(Command.DEACTIVATE_ELECTROMAGNET, pin_id_str)

    def enable_tcp(self, enable=True):
        """
        Enables or disables the TCP function (Tool Center Point).
        If activation is requested, the last recorded TCP value will be applied.
        The default value depends on the gripper equipped.
        If deactivation is requested, the TCP will be coincident with the tool_link.

        :param enable: True to enable, False otherwise.
        :type enable: Bool
        :rtype: None
        """
        self.__check_instance(enable, bool)
        self.__send_n_receive(Command.ENABLE_TCP, enable)

    def set_tcp(self, *args):
        """
        Activates the TCP function (Tool Center Point)
        and defines the transformation between the tool_link frame and the TCP frame.

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None
        """
        tcp_transform = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.SET_TCP, *tcp_transform)

    def reset_tcp(self):
        """
        Reset the TCP (Tool Center Point) transformation.
        The TCP will be reset according to the tool equipped.

        :rtype: None
        """
        self.__send_n_receive(Command.RESET_TCP)

    def tool_reboot(self):
        """
        Reboot the motor of the tool equipped. Useful when an Overload error occurs. (cf HardwareStatus)

        :rtype: None
        """
        self.__send_n_receive(Command.TOOL_REBOOT)

    # - Hardware

    def set_pin_mode(self, pin_id, pin_mode):
        """
        Set pin number pin_id to mode pin_mode

        :param pin_id:
        :type pin_id: PinID or str
        :param pin_mode:
        :type pin_mode: PinMode
        :rtype: None
        """
        self.__check_enum_belonging(pin_mode, PinMode)
        self.__check_instance(pin_id, (PinID, str))
        pin_id_str = pin_id.value if isinstance(pin_id, PinID) else pin_id

        self.__send_n_receive(Command.SET_PIN_MODE, pin_id_str, pin_mode)

    @property
    def digital_io_state(self):
        return self.get_digital_io_state()

    def get_digital_io_state(self):
        """
        Get Digital IO state : Names, modes, states.

        Example: ::

            digital_io_state = robot.digital_io_state
            digital_io_state = robot.get_digital_io_state()

        :return: List of DigitalPinObject instance
        :rtype: list[DigitalPinObject]
        """
        data = self.__send_n_receive(Command.GET_DIGITAL_IO_STATE)
        digital_pin_array = []
        for pin_info in data:
            name, mode, value = pin_info
            digital_pin_array.append(DigitalPinObject(PinID(name), name, PinMode(mode), PinState(bool(value))))
        return digital_pin_array

    def digital_write(self, pin_id, digital_state):
        """
        Set pin_id state to digital_state

        :param pin_id:
        :type pin_id: PinID or str
        :param digital_state:
        :type digital_state: PinState
        :rtype: None
        """
        self.__check_enum_belonging(digital_state, PinState)
        self.__check_instance(pin_id, (PinID, str))
        pin_id_str = pin_id.value if isinstance(pin_id, PinID) else pin_id

        self.__send_n_receive(Command.DIGITAL_WRITE, pin_id_str, digital_state)

    def digital_read(self, pin_id):
        """
        Read pin number pin_id and return its state

        :param pin_id:
        :type pin_id: PinID or str
        :rtype: PinState
        """
        self.__check_instance(pin_id, (PinID, str))
        pin_id_str = pin_id.value if isinstance(pin_id, PinID) else pin_id

        state_id = self.__send_n_receive(Command.DIGITAL_READ, pin_id_str)
        return PinState[state_id]

    @property
    def analog_io_state(self):
        return self.get_analog_io_state()

    def get_analog_io_state(self):
        """
        Get Analog IO state : Names, modes, states

        Example: ::

            analog_io_state = robot.analog_io_state
            analog_io_state = robot.get_analog_io_state()


        :return: List of AnalogPinObject instance
        :rtype: list[AnalogPinObject]
        """
        data = self.__send_n_receive(Command.GET_ANALOG_IO_STATE)

        analog_pin_array = []
        for pin_info in data:
            name, mode, value = pin_info
            analog_pin_array.append(AnalogPinObject(PinID(name), name, PinMode(mode), value))
        return analog_pin_array

    def analog_write(self, pin_id, value):
        """
        Set and analog pin_id state to a value

        :param pin_id:
        :type pin_id: PinID or str
        :param value: voltage between 0 and 5V
        :type value: float
        :rtype: None
        """
        self.__check_instance(pin_id, (PinID, str))
        self.__check_range_belonging(value, 0, 5)

        self.__send_n_receive(Command.ANALOG_WRITE, pin_id, value)

    def analog_read(self, pin_id):
        """
        Read the analog pin value

        :param pin_id:
        :type pin_id: PinID or str
        :rtype: float
        """
        self.__check_instance(pin_id, (PinID, str))

        return self.__send_n_receive(Command.ANALOG_READ, pin_id)

    @property
    def custom_button_state(self):
        return self.get_custom_button_state()

    def get_custom_button_state(self):
        """
        Get the Ned2's custom button state

        :return: True if pressed, False else
        :rtype: bool
        """
        return self.__send_n_receive(Command.CUSTOM_BUTTON_STATE)

    @property
    def hardware_status(self):
        return self.get_hardware_status()

    def get_hardware_status(self):
        """
        Get hardware status : Temperature, Hardware version, motors names & types ...

        :return: Infos contains in a HardwareStatusObject
        :rtype: HardwareStatusObject
        """
        data = self.__send_n_receive(Command.GET_HARDWARE_STATUS)

        rpi_temperature = float(data[0])
        hardware_version = data[1]
        connection_up = eval(data[2])
        error_message = data[3]
        calibration_needed = eval(data[4])
        calibration_in_progress = eval(data[5])

        motor_names = data[6]
        motor_types = data[7]
        temperatures = self.__map_list(data[8], float)
        voltages = self.__map_list(data[9], float)
        hardware_errors = data[10]

        hardware_status = HardwareStatusObject(rpi_temperature, hardware_version, connection_up, error_message,
                                               calibration_needed, calibration_in_progress,
                                               motor_names, motor_types,
                                               temperatures, voltages, hardware_errors)
        return hardware_status

    # - Conveyor

    def set_conveyor(self):
        """
        Activate a new conveyor and return its ID

        :return: New conveyor ID
        :rtype: ConveyorID
        """
        conveyor_id_str = self.__send_n_receive(Command.SET_CONVEYOR)
        conveyor_id = ConveyorID[conveyor_id_str]

        # If new conveyor has been found
        if conveyor_id != ConveyorID.NONE:
            return conveyor_id

        connected_conveyors_id = self.get_connected_conveyors_id()
        if not connected_conveyors_id:
            print("No conveyor connected !", file=sys.stderr)
            return ConveyorID.NONE
        else:
            print("No new conveyor detected, returning last connected conveyor", file=sys.stderr)
            return connected_conveyors_id[-1]

    def unset_conveyor(self, conveyor_id):
        """
        Remove specific conveyor.

        :param conveyor_id: Basically, ConveyorID.ONE or ConveyorID.TWO
        :type conveyor_id: ConveyorID
        """
        self.__send_n_receive(Command.UNSET_CONVEYOR, conveyor_id)

    def run_conveyor(self, conveyor_id, speed=50, direction=ConveyorDirection.FORWARD):
        """
        Run conveyor at id 'conveyor_id'

        :param conveyor_id:
        :type conveyor_id: ConveyorID
        :param speed:
        :type speed: int
        :param direction:
        :type direction: ConveyorDirection
        :rtype: None
        """
        self.control_conveyor(conveyor_id, control_on=True, speed=speed, direction=direction)

    def stop_conveyor(self, conveyor_id):
        """
        Stop conveyor at id 'conveyor_id'

        :param conveyor_id:
        :type conveyor_id: ConveyorID
        :rtype: None
        """
        self.control_conveyor(conveyor_id, control_on=False, speed=50, direction=ConveyorDirection.FORWARD)

    def control_conveyor(self, conveyor_id, control_on, speed, direction):
        """
        Control conveyor at id 'conveyor_id'

        :param conveyor_id:
        :type conveyor_id: ConveyorID
        :param control_on:
        :type control_on: bool
        :param speed: New speed which is a percentage of maximum speed
        :type speed: int
        :param direction: Conveyor direction
        :type direction: ConveyorDirection
        :rtype: None
        """
        self.__check_enum_belonging(conveyor_id, ConveyorID)
        self.__check_type(control_on, bool)
        self.__transform_to_type(speed, int)
        self.__check_enum_belonging(direction, ConveyorDirection)

        self.__send_n_receive(Command.CONTROL_CONVEYOR, conveyor_id, control_on, speed, direction)

    def get_connected_conveyors_id(self):
        """

        :return: List of the connected conveyors' ID
        :rtype: list[ConveyorID]
        """
        conveyors_list_str = self.__send_n_receive(Command.GET_CONNECTED_CONVEYORS_ID)
        conveyors_list = [ConveyorID[conveyor] for conveyor in conveyors_list_str]

        return conveyors_list

    # - Vision
    def get_img_compressed(self):
        """
        Get image from video stream in a compressed format. 
        Use ``uncompress_image`` from the vision package to uncompress it 

        :return: string containing a JPEG compressed image
        :rtype: str
        """
        _, img = self.__send_n_receive(Command.GET_IMAGE_COMPRESSED, with_payload=True)
        return img

    def set_brightness(self, brightness_factor):
        """
        Modify video stream brightness

        :param brightness_factor: How much to adjust the brightness. 0.5 will
            give a darkened image, 1 will give the original image while
            2 will enhance the brightness by a factor of 2.
        :type brightness_factor: float
        :rtype: None
        """
        self.__transform_to_type(brightness_factor, float)
        self.__check_range_belonging(brightness_factor, 0.0, np.Inf)
        self.__send_n_receive(Command.SET_IMAGE_BRIGHTNESS, brightness_factor)

    def set_contrast(self, contrast_factor):
        """
        Modify video stream contrast

        :param contrast_factor: While a factor of 1 gives original image.
            Making the factor towards 0 makes the image greyer, while factor>1 increases the contrast of the image.
        :type contrast_factor: float
        :rtype: None
        """
        self.__transform_to_type(contrast_factor, float)
        self.__check_range_belonging(contrast_factor, 0.0, np.Inf)
        self.__send_n_receive(Command.SET_IMAGE_CONTRAST, contrast_factor)

    def set_saturation(self, saturation_factor):
        """
        Modify video stream saturation

        :param saturation_factor: How much to adjust the saturation. 0 will
            give a black and white image, 1 will give the original image while
            2 will enhance the saturation by a factor of 2.
        :type saturation_factor: float
        :rtype: None
        """
        self.__transform_to_type(saturation_factor, float)
        self.__check_range_belonging(saturation_factor, 0.0, np.Inf)
        self.__send_n_receive(Command.SET_IMAGE_SATURATION, saturation_factor)

    def get_image_parameters(self):
        """
        Get last stream image parameters: Brightness factor, Contrast factor, Saturation factor.

        Brightness factor: How much to adjust the brightness. 0.5 will give a darkened image,
        1 will give the original image while 2 will enhance the brightness by a factor of 2.

        Contrast factor: A factor of 1 gives original image.
        Making the factor towards 0 makes the image greyer, while factor>1 increases the contrast of the image.

        Saturation factor: 0 will give a black and white image, 1 will give the original image while
        2 will enhance the saturation by a factor of 2.

        :return:  Brightness factor, Contrast factor, Saturation factor
        :rtype: float, float, float
        """
        brightness_factor, contrast_factor, saturation_factor = self.__send_n_receive(Command.GET_IMAGE_PARAMETERS)
        return brightness_factor, contrast_factor, saturation_factor

    def get_target_pose_from_rel(self, workspace_name, height_offset, x_rel, y_rel, yaw_rel):
        """
        Given a pose (x_rel, y_rel, yaw_rel) relative to a workspace, this function
        returns the robot pose in which the current tool will be able to pick an object at this pose.
        
        The height_offset argument (in m) defines how high the tool will hover over the workspace. If height_offset = 0,
        the tool will nearly touch the workspace.

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param x_rel: x relative pose (between 0 and 1)
        :type x_rel: float
        :param y_rel: y relative pose (between 0 and 1)
        :type y_rel: float
        :param yaw_rel: Angle in radians
        :type yaw_rel: float

        :return: target_pose
        :rtype: PoseObject
        """
        self.__check_type(workspace_name, str)
        self.__check_range_belonging(x_rel, 0.0, 1.0)
        self.__check_range_belonging(y_rel, 0.0, 1.0)

        height_offset, x_rel, y_rel, yaw_rel = self.__map_list([height_offset, x_rel, y_rel, yaw_rel], float)
        pose_array = self.__send_n_receive(Command.GET_TARGET_POSE_FROM_REL,
                                           workspace_name, height_offset, x_rel, y_rel, yaw_rel)
        pose_object = PoseObject(*pose_array)
        return pose_object

    def get_target_pose_from_cam(self, workspace_name, height_offset=0.0, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        First detects the specified object using the camera and then returns the robot pose in which the object can
        be picked with the current tool

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_pose, object_shape, object_color
        :rtype: (bool, PoseObject, ObjectShape, ObjectColor)
        """
        self.__check_type(workspace_name, str)
        height_offset = self.__transform_to_type(height_offset, float)
        self.__check_enum_belonging(shape, ObjectShape)
        self.__check_enum_belonging(color, ObjectColor)

        data_array = self.__send_n_receive(Command.GET_TARGET_POSE_FROM_CAM,
                                           workspace_name, height_offset, shape, color)
        obj_found = eval(data_array[0])
        if obj_found:
            pose_object = PoseObject(*data_array[1])
            shape_ret = data_array[2]
            color_ret = data_array[3]
        else:
            pose_object = PoseObject(0, 0, 0, 0, 0, 0)
            shape_ret = "ANY"
            color_ret = 'ANY'
        return obj_found, pose_object, ObjectShape[shape_ret], ObjectColor[color_ret]

    def __move_with_vision(self, command, workspace_name, height_offset, shape, color):
        self.__check_type(workspace_name, str)
        height_offset = self.__transform_to_type(height_offset, float)
        self.__check_enum_belonging(shape, ObjectShape)
        self.__check_enum_belonging(color, ObjectColor)

        data_array = self.__send_n_receive(command, workspace_name, height_offset, shape, color)

        obj_found = eval(data_array[0])
        if obj_found is True:
            shape_ret = data_array[1]
            color_ret = data_array[2]
        else:
            shape_ret = "ANY"
            color_ret = 'ANY'
        return obj_found, ObjectShape[shape_ret], ObjectColor[color_ret]

    def vision_pick(self, workspace_name, height_offset=0.0, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        Picks the specified object from the workspace. This function has multiple phases: \n
        | 1. detect object using the camera 
        | 2. prepare the current tool for picking 
        | 3. approach the object 
        | 4. move down to the correct picking pose 
        | 5. actuate the current tool 
        | 6. lift the object

        Example::

            robot = NiryoRobot(ip_address="x.x.x.x")
            robot.calibrate_auto()
            robot.move_pose(<observation_pose>)
            obj_found, shape_ret, color_ret = robot.vision_pick(<workspace_name>,
                                                                height_offset=0.0,
                                                                shape=ObjectShape.ANY,
                                                                color=ObjectColor.ANY)

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_shape, object_color
        :rtype: (bool, ObjectShape, ObjectColor)
        """
        return self.__move_with_vision(Command.VISION_PICK, workspace_name, height_offset, shape, color)

    def move_to_object(self, workspace_name, height_offset, shape, color):
        """
        Same as `get_target_pose_from_cam` but directly moves to this position

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_shape, object_color
        :rtype: (bool, ObjectShape, ObjectColor)
        """
        return self.__move_with_vision(Command.MOVE_TO_OBJECT, workspace_name, height_offset, shape, color)

    def detect_object(self, workspace_name, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        Detect object in workspace and return its pose and characteristics

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_pose, object_shape, object_color
        :rtype: (bool, PoseObject, str, str)
        """
        self.__check_type(workspace_name, str)
        self.__check_enum_belonging(shape, ObjectShape)
        self.__check_enum_belonging(color, ObjectColor)

        data_array = self.__send_n_receive(Command.DETECT_OBJECT, workspace_name, shape, color)
        obj_found = eval(data_array[0])
        if not obj_found:
            rel_pose_array = 3 * [0.0]
            shape = "ANY"
            color = "ANY"
        else:
            rel_pose_array = data_array[1:4]
            shape = data_array[4]
            color = data_array[5]

        return obj_found, rel_pose_array, ObjectShape[shape], ObjectColor[color]

    def get_camera_intrinsics(self):
        """
        Get calibration object: camera intrinsics, distortions coefficients

        :return: camera intrinsics, distortions coefficients
        :rtype: (list[list[float]], list[list[float]])
        """
        data = self.__send_n_receive(Command.GET_CAMERA_INTRINSICS)

        mtx = np.reshape(data[0], (3, 3))
        dist = np.expand_dims(data[1], axis=0)
        return mtx, dist

    # - Workspace
    def save_workspace_from_robot_poses(self, workspace_name, pose_origin, pose_2, pose_3, pose_4):
        """
        Save workspace by giving the poses of the robot to point its 4 corners
        with the calibration Tip. Corners should be in the good order.
        Markers' pose will be deduced from these poses

        Poses should be either a list [x, y, z, roll, pitch, yaw] or a PoseObject

        :param workspace_name: workspace name, maximum lenght 30 char.
        :type workspace_name: str
        :param pose_origin:
        :type pose_origin: Union[list[float], PoseObject]
        :param pose_2:
        :type pose_2: Union[list[float], PoseObject]
        :param pose_3:
        :type pose_3: Union[list[float], PoseObject]
        :param pose_4:
        :type pose_4: Union[list[float], PoseObject]

        :rtype: None
        """
        self.__check_type(workspace_name, str)

        param_list = [workspace_name]
        for pose in (pose_origin, pose_2, pose_3, pose_4):
            pose_list = self.__args_pose_to_list(pose)
            param_list.append(pose_list)
        self.__send_n_receive(Command.SAVE_WORKSPACE_FROM_POSES, *param_list)

    def save_workspace_from_points(self, workspace_name, point_origin, point_2, point_3, point_4):
        """
        Save workspace by giving the points of worskpace's 4 corners. Points are written as [x, y, z]
        Corners should be in the good order.

        :param workspace_name: workspace name, maximum lenght 30 char.
        :type workspace_name: str
        :param point_origin:
        :type point_origin: list[float]
        :param point_2:
        :type point_2: list[float]
        :param point_3:
        :type point_3: list[float]
        :param point_4:
        :type point_4: list[float]
        :rtype: None
        """
        self.__check_type(workspace_name, str)

        param_list = [workspace_name]
        for point in (point_origin, point_2, point_3, point_4):
            param_list.append(self.__map_list(point, float))
        self.__send_n_receive(Command.SAVE_WORKSPACE_FROM_POINTS, *param_list)

    def delete_workspace(self, workspace_name):
        """
        Delete workspace from robot's memory

        :param workspace_name:
        :type workspace_name: str
        :rtype: None
        """
        self.__check_type(workspace_name, str)

        self.__send_n_receive(Command.DELETE_WORKSPACE, workspace_name)

    def get_workspace_poses(self, workspace_name):
        raise NotImplementedError

    def get_workspace_ratio(self, workspace_name):
        """
        Get workspace ratio from robot's memory

        :param workspace_name:
        :type workspace_name: str
        :rtype: float
        """
        self.__check_type(workspace_name, str)
        return self.__send_n_receive(Command.GET_WORKSPACE_RATIO, workspace_name)

    def get_workspace_list(self):
        """
        Get list of workspaces' name store in robot's memory

        :rtype: list[str]
        """
        return self.__send_n_receive(Command.GET_WORKSPACE_LIST)

    # Sound

    def get_sounds(self):
        """
        Get sound name list

        :return: list of the sounds of the robot
        :rtype: list[string]
        """
        return self.__send_n_receive(Command.GET_SOUNDS)

    def play_sound(self, sound_name, wait_end=True, start_time_sec=0, end_time_sec=0):
        """
        Play a sound from the robot

        :param sound_name: Name of the sound to play
        :type sound_name: string
        :param wait_end: wait for the end of the sound before exiting the function
        :type wait_end: bool
        :param start_time_sec: start the sound from this value in seconds
        :type start_time_sec: float
        :param end_time_sec: end the sound at this value in seconds
        :type end_time_sec: float
        :rtype: None
        """
        self.__check_list_belonging(sound_name, self.get_sounds())
        self.__send_n_receive(Command.PLAY_SOUND, sound_name, wait_end, start_time_sec, end_time_sec)

    def set_volume(self, sound_volume):
        """
        Set the volume percentage of the robot.

        :param sound_volume: volume percentage of the sound (0: no sound, 100: max sound)
        :type sound_volume: int
        :rtype: None
        """
        self.__check_range_belonging(sound_volume, 0, 200)
        self.__send_n_receive(Command.SET_VOLUME, sound_volume)

    def stop_sound(self):
        """
        Stop a sound being played.

        :rtype: None
        """
        self.__send_n_receive(Command.STOP_SOUND)

    def get_sound_duration(self, sound_name):
        """
        Returns the duration in seconds of a sound stored in the robot database
        raise SoundRosWrapperException if the sound doesn't exists

        :param sound_name: name of sound
        :type sound_name: string
        :return: sound duration in seconds
        :rtype: float
        """
        self.__check_list_belonging(sound_name, self.get_sounds())
        return self.__send_n_receive(Command.GET_SOUND_DURATION, sound_name)

    def say(self, text, language=0):
        """
        Use gtts (Google Text To Speech) to interprete a string as sound
        Languages available are:
        * English: 0
        * French: 1
        * Spanish: 2
        * Mandarin: 3
        * Portuguese: 4

        Example ::

            robot.say("Hello", 0)
            robot.say("Bonjour", 1)
            robot.say("Hola", 2)

        :param text: Text that needs to be spoken < 100 char
        :type text: string
        :param language: language of the text
        :type language: int
        :rtype: None
        """
        self.__send_n_receive(Command.SAY, text, language)

    # Led Ring

    def set_led_color(self, led_id, color):
        """
        Lights up an LED in one colour. RGB colour between 0 and 255.

        Example: ::

            robot.set_led_color(5, [15, 50, 255])

        :param led_id: Id of the led: between 0 and 29
        :type led_id: int
        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_SET_LED, led_id, color)

    def led_ring_solid(self, color):
        """
        Set the whole Led Ring to a fixed color.

        Example: ::

            robot.led_ring_solid([15, 50, 255])

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_SOLID, color, True)

    def led_ring_turn_off(self):
        """
        Turn off all LEDs

        Example: ::

            robot.led_ring_turn_off()

        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_TURN_OFF, True)

    def led_ring_flashing(self, color, period=0, iterations=0, wait=False):
        """
        Flashes a color according to a frequency. The frequency is equal to 1 / period.

        Examples: ::

            robot.led_ring_flashing([15, 50, 255])
            robot.led_ring_flashing([15, 50, 255], 1, 100, True)
            robot.led_ring_flashing([15, 50, 255], iterations=20, wait=True)

            frequency = 20  # Hz
            total_duration = 10 # seconds
            robot.flashing([15, 50, 255], 1./frequency, total_duration * frequency , True)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive flashes. If 0, the Led Ring flashes endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish all iterations or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_FLASH, color, period, iterations, wait)

    def led_ring_alternate(self, color_list, period=0, iterations=0, wait=False):
        """
        Several colors are alternated one after the other.

        Examples: ::

            color_list = [
                [15, 50, 255],
                [255, 0, 0],
                [0, 255, 0],
            ]

            robot.led_ring_alternate(color_list)
            robot.led_ring_alternate(color_list, 1, 100, True)
            robot.led_ring_alternate(color_list, iterations=20, wait=True)

        :param color_list: Led color list of lists of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color_list: list[list[float]]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive alternations. If 0, the Led Ring alternates endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish all iterations or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_ALTERNATE, color_list, period, iterations, wait)

    def led_ring_chase(self, color, period=0, iterations=0, wait=False):
        """
        Movie theater light style chaser animation.

        Examples: ::


            robot.led_ring_chase([15, 50, 255])
            robot.led_ring_chase([15, 50, 255], 1, 100, True)
            robot.led_ring_chase([15, 50, 255], iterations=20, wait=True)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive chase. If 0, the animation continues endlessly.
            One chase just lights one Led every 3 LEDs.
        :type iterations: int
        :param wait: The service wait for the animation to finish all iterations or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_CHASE, color, period, iterations, wait)

    def led_ring_wipe(self, color, period=0, wait=False):
        """
        Wipe a color across the Led Ring, light a Led at a time.

        Examples: ::

            robot.led_ring_wipe([15, 50, 255])
            robot.led_ring_wipe([15, 50, 255], 1, True)
            robot.led_ring_wipe([15, 50, 255], wait=True)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param wait: The service wait for the animation to finish or not to answer.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_WIPE, color, period, wait)

    def led_ring_rainbow(self, period=0, iterations=0, wait=False):
        """
        Draw rainbow that fades across all LEDs at once.

        Examples: ::

            robot.led_ring_rainbow()
            robot.led_ring_rainbow(5, 2, True)
            robot.led_ring_rainbow(wait=True)

        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive rainbows. If 0, the animation continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_RAINBOW, period, iterations, wait)

    def led_ring_rainbow_cycle(self, period=0, iterations=0, wait=False):
        """
        Draw rainbow that uniformly distributes itself across all LEDs.

        Examples: ::

            robot.led_ring_rainbow_cycle()
            robot.led_ring_rainbow_cycle(5, 2, True)
            robot.led_ring_rainbow_cycle(wait=True)

        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive rainbow cycles. If 0, the animation continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_RAINBOW_CYCLE, period, iterations, wait)

    def led_ring_rainbow_chase(self, period=0, iterations=0, wait=False):
        """
        Rainbow chase animation, like the led_ring_chase method.

        Examples: ::

            robot.led_ring_rainbow_chase()
            robot.led_ring_rainbow_chase(5, 2, True)
            robot.led_ring_rainbow_chase(wait=True)

        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive rainbow cycles. If 0, the animation continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_RAINBOW_CHASE, period, iterations, wait)

    def led_ring_go_up(self, color, period=0, iterations=0, wait=False):
        """
        LEDs turn on like a loading circle, and are then all turned off at once.

        Examples: ::

            robot.led_ring_go_up([15, 50, 255])
            robot.led_ring_go_up([15, 50, 255], 1, 100, True)
            robot.led_ring_go_up([15, 50, 255], iterations=20, wait=True)


        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive turns around the Led Ring. If 0, the animation
            continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_GO_UP, color, period, iterations, wait)

    def led_ring_go_up_down(self, color, period=0, iterations=0, wait=False):
        """
        LEDs turn on like a loading circle, and are turned off the same way.

        Examples: ::

            robot.led_ring_go_up_down([15, 50, 255])
            robot.led_ring_go_up_down([15, 50, 255], 1, 100, True)
            robot.led_ring_go_up_down([15, 50, 255], iterations=20, wait=True)


        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive turns around the Led Ring. If 0, the animation
            continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_GO_UP_DOWN, color, period, iterations, wait)

    def led_ring_breath(self, color, period=0, iterations=0, wait=False):
        """
        Variation of the light intensity of the LED ring, similar to human breathing.

        Examples: ::

            robot.led_ring_breath([15, 50, 255])
            robot.led_ring_breath([15, 50, 255], 1, 100, True)
            robot.led_ring_breath([15, 50, 255], iterations=20, wait=True)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive turns around the Led Ring. If 0, the animation
            continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_BREATH, color, period, iterations, wait)

    def led_ring_snake(self, color, period=0, iterations=0, wait=False):
        """
        A small coloured snake (certainly a python :D ) runs around the LED ring.

        Examples: ::

            robot.led_ring_snake([15, 50, 255])
            robot.led_ring_snake([15, 50, 255], 1, 100, True)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default duration will be used.
        :type period: float
        :param iterations: Number of consecutive turns around the Led Ring. If 0, the animation
            continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_SNAKE, color, period, iterations, wait)

    def led_ring_custom(self, led_colors):
        """
        Sends a colour command to all LEDs of the LED ring.
        The function expects a list of colours for the 30 LEDs  of the robot.

        Example: ::

            led_list = [[i / 30. * 255 , 0, 255 - i / 30.] for i in range(30)]
            robot.led_ring_custom(led_list)

        :param led_colors: List of size 30 of led color in a list of size 3[R, G, B].
                RGB channels from 0 to 255.
        :type led_colors: list[list[float]]
        :return: status, message
        :rtype: (int, str)
        """
        return self.__send_n_receive(Command.LED_RING_CUSTOM, led_colors)



"""
This file shows how to use PyNiryo to do a simple pick and place
with Ned using a conveyor
"""

# Imports
from pyniryo import *

# -- MUST Change these variables
simulation_mode = True
gripper_used = ToolID.GRIPPER_2  # Tool used for picking
gpio_sensor = PinID.GPIO_1A  # Pin of the sensor
# Set robot address
robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simulation = "127.0.0.1"
robot_ip_address = robot_ip_address_simulation if simulation_mode else robot_ip_address_rpi

# -- Should Change these variables
# The pick pose
pick_pose = PoseObject(
    x=0.25, y=0., z=0.14,
    roll=-0., pitch=1.57, yaw=0.0,
)
# The Place pose
place_pose = PoseObject(
    x=-0.01, y=-0.23, z=0.12,
    roll=-0., pitch=1.57, yaw=-1.57)


def pick_n_place_w_conveyor(niyro_robot):
    # Enable connection with conveyor
    conveyor_id = niyro_robot.set_conveyor()
    # Turn conveyor on
    niyro_robot.control_conveyor(conveyor_id=conveyor_id, control_on=True,
                                 speed=50, direction=ConveyorDirection.FORWARD)
    # Wait for sensor to turn to low state which means it has something in front of it
    while not niyro_robot.digital_read(gpio_sensor) == PinState.LOW:
        niyro_robot.wait(0.1)
    # Turn conveyor off
    niyro_robot.control_conveyor(conveyor_id=conveyor_id, control_on=False,
                                 speed=0, direction=ConveyorDirection.FORWARD)
    # Pick
    niyro_robot.pick_from_pose(pick_pose)
    # Place
    niyro_robot.place_from_pose(place_pose)


# -- MAIN PROGRAM

if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Calibrate robot if robot needs calibration
    robot.calibrate_auto()
    # Update tool
    robot.update_tool()
    # Launching main process
    pick_n_place_w_conveyor(robot)
    # Releasing connection
    robot.close_connection()



from enum import Enum, unique

TCP_PORT = 40001  # Port used for TCP communication
READ_SIZE = 512  # Buffer used for receiving TCP packet
TCP_TIMEOUT = 5  # Time before timeout is raised

# - Infos on the JSON size
# nbr_bytes : number of bytes on which the size is coded
# type : size coding type
DEFAULT_PACKET_SIZE_INFOS = {
    "nbr_bytes": 2,
    "type": '@H',
}


@unique
class CalibrateMode(Enum):
    """
    Enumeration of Calibration Modes
    """
    AUTO = 0
    MANUAL = 1


@unique
class RobotAxis(Enum):
    """
    Enumeration of Robot Axis : it used for Shift command
    """
    X = 0
    Y = 1
    Z = 2
    ROLL = 3
    PITCH = 4
    YAW = 5


@unique
class ToolID(Enum):
    """
    Enumeration of Tools IDs
    """
    NONE = 0
    GRIPPER_1 = 11
    GRIPPER_2 = 12
    GRIPPER_3 = 13
    ELECTROMAGNET_1 = 30
    VACUUM_PUMP_1 = 31


@unique
class PinMode(Enum):
    """
    Enumeration of Pin Modes
    """
    OUTPUT = 0
    INPUT = 1


@unique
class PinState(Enum):
    """
    Pin State is either LOW or HIGH
    """
    LOW = False
    HIGH = True


@unique
class PinID(Enum):
    """
    Enumeration of Robot Pins
    """
    GPIO_1A = "1A"
    GPIO_1B = "1B"
    GPIO_1C = "1C"
    GPIO_2A = "2A"
    GPIO_2B = "2B"
    GPIO_2C = "2C"

    SW_1 = "SW1"
    SW_2 = "SW2"

    DO1 = "DO1"
    DO2 = "DO2"
    DO3 = "DO3"
    DO4 = "DO4"
    DI1 = "DI1"
    DI2 = "DI2"
    DI3 = "DI3"
    DI4 = "DI4"
    DI5 = "DI5"

    AI1 = "AI1"
    AI2 = "AI2"
    AO1 = "AO1"
    AO2 = "AO2"


# - Conveyor
class ConveyorID(Enum):
    """
    Enumeration of Conveyor IDs used for Conveyor control
    """
    NONE = 0
    ID_1 = 12
    ID_2 = 13
    ID_3 = 9
    ID_4 = 10


@unique
class ConveyorDirection(Enum):
    """
    Enumeration of Conveyor Directions used for Conveyor control
    """
    FORWARD = 1
    BACKWARD = -1


# - Vision
@unique
class ObjectColor(Enum):
    """
    Enumeration of Colors available for image processing
    """
    RED = "RED"
    BLUE = "BLUE"
    GREEN = "GREEN"
    ANY = "ANY"


@unique
class ObjectShape(Enum):
    """
    Enumeration of Shapes available for image processing
    """
    SQUARE = "SQUARE"
    CIRCLE = "CIRCLE"
    ANY = "ANY"


@unique
class Command(Enum):
    """
    Enumeration of all commands used
    """

    # Main purpose
    CALIBRATE = 0
    SET_LEARNING_MODE = 1
    GET_LEARNING_MODE = 2
    SET_ARM_MAX_VELOCITY = 3
    SET_JOG_CONTROL = 4

    # - Move
    # Pose

    GET_JOINTS = 10
    GET_POSE = 11
    GET_POSE_QUAT = 12

    MOVE_JOINTS = 20
    MOVE_POSE = 21
    SHIFT_POSE = 22

    MOVE_LINEAR_POSE = 23
    SHIFT_LINEAR_POSE = 24

    JOG_JOINTS = 25
    JOG_POSE = 26

    FORWARD_KINEMATICS = 27
    INVERSE_KINEMATICS = 28

    # Saved Pose
    GET_POSE_SAVED = 50
    SAVE_POSE = 51
    DELETE_POSE = 52
    GET_SAVED_POSE_LIST = 53

    # Pick & Place

    PICK_FROM_POSE = 60
    PLACE_FROM_POSE = 61
    PICK_AND_PLACE = 62

    # Trajectories
    GET_TRAJECTORY_SAVED = 80
    EXECUTE_TRAJECTORY_FROM_POSES = 81
    EXECUTE_TRAJECTORY_SAVED = 82
    SAVE_TRAJECTORY = 83
    DELETE_TRAJECTORY = 84
    GET_SAVED_TRAJECTORY_LIST = 85
    EXECUTE_TRAJECTORY_FROM_POSES_AND_JOINTS = 86

    # - Tools
    UPDATE_TOOL = 120
    OPEN_GRIPPER = 121
    CLOSE_GRIPPER = 122
    PULL_AIR_VACUUM_PUMP = 123
    PUSH_AIR_VACUUM_PUMP = 124
    SETUP_ELECTROMAGNET = 125
    ACTIVATE_ELECTROMAGNET = 126
    DEACTIVATE_ELECTROMAGNET = 127
    GET_CURRENT_TOOL_ID = 128
    GRASP_WITH_TOOL = 129
    RELEASE_WITH_TOOL = 130
    ENABLE_TCP = 140
    SET_TCP = 141
    RESET_TCP = 142
    TOOL_REBOOT = 145

    # - Hardware
    SET_PIN_MODE = 150
    DIGITAL_WRITE = 151
    DIGITAL_READ = 152
    GET_DIGITAL_IO_STATE = 153
    GET_HARDWARE_STATUS = 154
    ANALOG_WRITE = 155
    ANALOG_READ = 156
    GET_ANALOG_IO_STATE = 157
    CUSTOM_BUTTON_STATE = 158

    # - Conveyor
    SET_CONVEYOR = 180
    UNSET_CONVEYOR = 181
    CONTROL_CONVEYOR = 182
    GET_CONNECTED_CONVEYORS_ID = 183

    # - Vision
    GET_IMAGE_COMPRESSED = 200
    GET_TARGET_POSE_FROM_REL = 201
    GET_TARGET_POSE_FROM_CAM = 202

    VISION_PICK = 203
    MOVE_TO_OBJECT = 205
    DETECT_OBJECT = 204

    GET_CAMERA_INTRINSICS = 210

    SAVE_WORKSPACE_FROM_POSES = 220
    SAVE_WORKSPACE_FROM_POINTS = 221
    DELETE_WORKSPACE = 222
    GET_WORKSPACE_RATIO = 223
    GET_WORKSPACE_LIST = 224

    SET_IMAGE_BRIGHTNESS = 230
    SET_IMAGE_CONTRAST = 231
    SET_IMAGE_SATURATION = 232
    GET_IMAGE_PARAMETERS = 235

    # - Sound
    PLAY_SOUND = 240
    SET_VOLUME = 241
    STOP_SOUND = 242
    DELETE_SOUND = 243
    IMPORT_SOUND = 244
    GET_SOUNDS = 245
    GET_SOUND_DURATION = 246
    SAY = 247

    # Led Ring
    LED_RING_SOLID = 250
    LED_RING_TURN_OFF = 251
    LED_RING_FLASH = 252
    LED_RING_ALTERNATE = 253
    LED_RING_CHASE = 254
    LED_RING_WIPE = 255
    LED_RING_RAINBOW = 256
    LED_RING_RAINBOW_CYCLE = 257
    LED_RING_RAINBOW_CHASE = 258
    LED_RING_GO_UP = 259
    LED_RING_GO_UP_DOWN = 260
    LED_RING_BREATH = 261
    LED_RING_SNAKE = 262
    LED_RING_CUSTOM = 263
    LED_RING_SET_LED = 264



import numpy as np
import cv2
from .math_functions import euclidean_dist_2_pts

IM_EXTRACT_SMALL_SIDE_PIXELS = 200


def extract_img_markers(img, workspace_ratio=1.0):
    """
    Extract working area from an image thanks to 4 Niryo's markers
    :param img: OpenCV image which contain 4 Niryo's markers
    :param workspace_ratio: Ratio between the width and the height of the area represented by the markers
    :return: extracted and warped working area image
    """

    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    img_thresh = cv2.adaptiveThreshold(gray, maxValue=255, adaptiveMethod=cv2.ADAPTIVE_THRESH_MEAN_C,
                                       thresholdType=cv2.THRESH_BINARY, blockSize=15, C=25)

    list_good_candidates = find_markers_from_img_thresh(img_thresh)
    if not list_good_candidates or len(list_good_candidates) > 6:
        return None

    if len(list_good_candidates) == 4:
        list_good_candidates = sort_markers_detection(list_good_candidates)
    else:
        list_good_candidates = complicated_sort_markers(list_good_candidates, workspace_ratio=workspace_ratio)
        if list_good_candidates is None:
            return None

    im_cut = extract_sub_img(img, list_good_candidates, ratio_w_h=workspace_ratio)
    return im_cut


def extract_sub_img(img, list_corners, ratio_w_h=1.0):
    """
    Extract an small image from a big one using a Perspective Warp
    :param img: Big image from which the small one will be extracted
    :param list_corners: corners list of the small image
    :param ratio_w_h: Width over Height ratio of the area. It helps to not stretch the working area image
    :return: extracted and warped image
    """
    if list_corners is None or len(list_corners) != 4:
        return None

    if ratio_w_h >= 1.0:
        target_w_area = int(round(ratio_w_h * IM_EXTRACT_SMALL_SIDE_PIXELS))
        target_h_area = IM_EXTRACT_SMALL_SIDE_PIXELS
    else:
        ratio_w_h = 1.0 / ratio_w_h
        target_h_area = int(round(ratio_w_h * IM_EXTRACT_SMALL_SIDE_PIXELS))
        target_w_area = IM_EXTRACT_SMALL_SIDE_PIXELS

    points_grid = []

    for marker in list_corners:
        points_grid.append(marker.get_center())
    points_grid = np.array(points_grid, dtype=np.float32)
    final_pts = np.array(
        [[0, 0], [target_w_area - 1, 0],
         [target_w_area - 1, target_h_area - 1], [0, target_h_area - 1]],
        dtype=np.float32)
    transfo_matrix = cv2.getPerspectiveTransform(points_grid, final_pts)
    # print transfo_matrix
    # print np.linalg.det(transfo_matrix)
    area_im = cv2.warpPerspective(img, transfo_matrix, (target_w_area, target_h_area))
    return area_im


def draw_markers(img, workspace_ratio=1.0):
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    img_thresh = cv2.adaptiveThreshold(gray, maxValue=255, adaptiveMethod=cv2.ADAPTIVE_THRESH_MEAN_C,
                                       thresholdType=cv2.THRESH_BINARY, blockSize=15, C=32)

    list_good_candidates = find_markers_from_img_thresh(img_thresh)
    if not list_good_candidates:
        return False, img
    im_draw = img.copy()
    for marker in list_good_candidates:
        cx, cy = marker.get_center()
        radius = marker.get_radius()
        cv2.circle(im_draw, (cx, cy), radius, (0, 0, 255), 2)
    if len(list_good_candidates) > 6:
        return False, im_draw

    if len(list_good_candidates) == 4:
        list_good_candidates = sort_markers_detection(list_good_candidates)
    else:
        list_good_candidates = complicated_sort_markers(list_good_candidates, workspace_ratio=workspace_ratio)
        if list_good_candidates is None:
            return False, im_draw

    for i, marker in enumerate(list_good_candidates[:4]):
        cx, cy = marker.get_center()
        radius = marker.get_radius()
        cv2.circle(im_draw, (cx, cy), radius, (0, 200, 0), 2)
        cv2.putText(im_draw, "{}".format(i + 1),
                    (cx + 5, cy - 15), cv2.FONT_HERSHEY_SIMPLEX, 0.9, (0, 0, 0), 3)
        cv2.putText(im_draw, "{}".format(i + 1),
                    (cx + 5, cy - 15), cv2.FONT_HERSHEY_SIMPLEX, 0.9, (0, 200, 0), 2)
    return True, im_draw


class PotentialMarker:
    def __init__(self, center, radius, cnt):
        self.center = center
        self.x = center[0]
        self.y = center[1]
        self.radius = radius
        self.contour = cnt
        self.is_merged = False

    def get_center(self):
        return self.center

    def __str__(self):
        return "{} - {} - {}".format(self.x, self.y, self.radius)

    def __repr__(self):
        return self.__str__()


class Marker:
    def __init__(self, potential_marker):
        self.list_centers = [potential_marker.get_center()]
        self.list_radius = [potential_marker.radius]
        self.list_contours = [potential_marker.contour]
        self.cx = self.list_centers[0][0]
        self.cy = self.list_centers[0][1]
        self.radius = potential_marker.radius
        self.identifiant = None
        self.value_for_id = None

    def get_radius(self):
        return self.radius

    def get_center(self):
        return self.cx, self.cy

    def add_circle(self, obj_potential_marker):
        self.list_centers.append(obj_potential_marker.get_center())
        self.list_radius.append(obj_potential_marker.radius)
        obj_potential_marker.is_merged = True

        (x, y) = np.mean(self.list_centers, axis=0)
        self.cx, self.cy = int(round(x)), int(round(y))

        self.radius = int(round(max(self.list_radius)))

    def nb_circles(self):
        return len(self.list_centers)

    def get_id_from_slice(self, img_thresh):
        x, y, w, h = self.cx - 1, self.cy - 1, 3, 3
        self.value_for_id = np.mean(img_thresh[y:y + h, x:x + w])
        # return value_for_id
        if self.value_for_id > 200:
            self.identifiant = "A"
        else:
            self.identifiant = "B"

        return self.identifiant
        # return value_for_id

    def __str__(self):
        return "{} - {}".format(self.nb_circles(), self.list_centers)

    def __repr__(self):
        return self.__str__()


def sort_markers_detection(list_markers):
    def rotate_list(list_, n):
        return list_[n:] + list_[:n]

    list_sort_y = sorted(list_markers, key=lambda m: m.cy)

    top1, top2, bottom1, bottom2 = list_sort_y

    if top1.cx < top2.cx:
        top_left = top1
        top_right = top2
    else:
        top_left = top2
        top_right = top1

    if bottom1.cx < bottom2.cx:
        bottom_left = bottom1
        bottom_right = bottom2
    else:
        bottom_left = bottom2
        bottom_right = bottom1

    list_markers_unsorted = [top_left, top_right, bottom_right, bottom_left]
    list_id = [marker.identifiant for marker in list_markers_unsorted]

    if list_id.count("A") == 1:
        list_corners_sorted = rotate_list(list_markers_unsorted, n=list_id.index("A"))
    elif list_id.count("B") == 1:
        list_corners_sorted = rotate_list(list_markers_unsorted, n=list_id.index("B"))
    else:
        return list_markers_unsorted

    return list_corners_sorted


def complicated_sort_markers(list_markers, workspace_ratio):
    import itertools

    if workspace_ratio >= 1.0:
        target_w_area = int(round(workspace_ratio * 200))
        target_h_area = 200
    else:
        ratio_w_h = 1.0 / workspace_ratio
        target_h_area = int(round(ratio_w_h * 200))
        target_w_area = 200
    list_id = [marker.identifiant for marker in list_markers]
    count_type_a = list_id.count("A")
    count_type_b = list_id.count("B")
    if count_type_a < 3 > count_type_b:
        return None
    if count_type_a < count_type_b:
        id_first_marker = "A"
        id_second_marker = "B"
    else:
        id_first_marker = "B"
        id_second_marker = "A"
    list_combinaisons = []
    list_marker_1 = [marker for marker in list_markers if marker.identifiant == id_first_marker]
    list_marker_2 = [marker for marker in list_markers if marker.identifiant == id_second_marker]
    if list_marker_1:
        list_combinaisons_marker_2 = itertools.combinations(list_marker_2, 3)
        for marker1 in list_marker_1:
            for combi_markers2 in list_combinaisons_marker_2:
                combin = [marker1] + list(combi_markers2)

                list_combinaisons.append(sort_markers_detection(combin))
    else:
        for combinaison in itertools.combinations(list_marker_2, 4):
            list_combinaisons.append(combinaison)
    if not list_combinaisons:
        return None

    final_pts = np.array(
        [[0, 0], [target_w_area - 1, 0],
         [target_w_area - 1, target_h_area - 1], [0, target_h_area - 1]],
        dtype=np.float32)
    list_det_transfo_matrix = []
    for combin in list_combinaisons:
        points_grid = np.array([[mark.cx, mark.cy] for mark in combin], dtype=np.float32)

        transfo_matrix = cv2.getPerspectiveTransform(points_grid, final_pts)
        list_det_transfo_matrix.append(np.linalg.det(transfo_matrix))

    best_combin_ind = np.argmin(abs(np.array(list_det_transfo_matrix) - 1))
    best_markers = list_combinaisons[best_combin_ind]
    return best_markers


def find_markers_from_img_thresh(img_thresh, max_dist_between_centers=3, min_radius_circle=4,
                                 max_radius_circle=35, min_radius_marker=7):
    contours = cv2.findContours(img_thresh, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)[-2]
    list_potential_markers = []
    for cnt in contours:
        (x, y), radius = cv2.minEnclosingCircle(cnt)
        if not min_radius_circle < radius < max_radius_circle:
            continue
        center = (int(round(x)), int(round(y)))
        radius = int(radius)
        list_potential_markers.append(PotentialMarker(center, radius, cnt))

    list_potential_markers = sorted(list_potential_markers, key=lambda m: m.x)
    list_good_candidates = []

    for i, potential_marker in enumerate(list_potential_markers):
        if potential_marker.is_merged:
            continue
        marker1 = Marker(potential_marker)
        center_marker = marker1.get_center()

        for potential_marker2 in list_potential_markers[i + 1:]:
            if potential_marker.is_merged:
                continue
            center_potential = potential_marker2.get_center()
            if center_potential[0] - center_marker[0] > max_dist_between_centers:
                break
            dist = euclidean_dist_2_pts(center_marker, center_potential)
            if dist <= max_dist_between_centers:
                marker1.add_circle(potential_marker2)
                center_marker = marker1.get_center()

        if marker1.nb_circles() > 2 and marker1.radius >= min_radius_marker:
            list_good_candidates.append(marker1)
            marker1.get_id_from_slice(img_thresh)

    return list_good_candidates



"""
This script shows an example of how to use Ned's vision with 2
workspaces : 1 for picking & 1 for packing

The robot will firstly detect one pack place, then it will take one object
from the picking area. This object will be pack and then, the pack will be
place on the side of the working area
"""

from pyniryo import *

# -- MUST Change these variables
robot_ip_address = "10.10.10.10"  # IP address of Ned
workspace_pick_name = "workspace_1"  # Robot's picking Workspace Name
workspace_place_name = "workspace_place"  # Robot's placing Workspace Name

# -- Should Change these variables
# The pose from where the image processing for picking happens
observation_pose_wks_pick = PoseObject(
    x=0.20, y=0., z=0.3,
    roll=0.0, pitch=1.57, yaw=0.0,
)
# The pose from where the image processing for placing happens
observation_pose_wks_place = PoseObject(
    x=0.00, y=-0.2, z=0.35,
    roll=0.0, pitch=1.57, yaw=-1.57,
)
# First conditioning pose in the cubes row
pos_conditioning1 = PoseObject(
    x=0.1, y=-0.29, z=0.12,
    roll=-0., pitch=1.57, yaw=-1.57
)
# First conditioning pose in the circles row
pos_conditioning2 = PoseObject(
    x=-0.145, y=-0.282, z=0.12,
    roll=-0., pitch=1.57, yaw=-1.57
)

# Joints where the robot goes at the end of its process
sleep_joints = [0.0, 0.55, -1.2, 0.0, 0.0, 0.0]


def process(niryo_robot):
    catch_count_category_1 = 0
    catch_count_category_2 = 0
    try_without_success = 0
    # Loop until too much failures or 6 objects have been already picked
    while try_without_success < 3 and catch_count_category_1 + catch_count_category_2 < 6:
        obj_pose = None
        # Moving to observation pose over the packing workspace
        niryo_robot.move_pose(*observation_pose_wks_place.to_list())
        # Trying to get place pose from python ROS wrapper
        ret = niryo_robot.get_target_pose_from_cam(workspace_place_name,
                                                   height_offset=0.0,
                                                   shape=ObjectShape.ANY,
                                                   color=ObjectColor.ANY)
        # Unpacking return result
        status, obj_found, place_pose, shape, color = ret
        if not status or not obj_found:  # Aborting iteration if issue
            try_without_success += 1
            continue

        # Moving to observation pose over the picking workspace
        niryo_robot.move_pose(*observation_pose_wks_pick.to_list())

        try_without_success_within = 0
        obj_found_within = False
        should_leave = False
        shape_obj = None
        while not obj_found_within:  # Looping over picking area
            ret = niryo_robot.get_target_pose_from_cam(workspace_pick_name,
                                                       height_offset=0.0,
                                                       shape=ObjectShape.ANY,
                                                       color=ObjectColor.ANY)
            status, obj_found_within, obj_pose, shape_obj, color_obj = ret
            if not status or not obj_found_within:
                try_without_success_within += 1
                if try_without_success_within > 3:
                    should_leave = True
                    break
                continue
        if should_leave:  # If nothing has been found, nothing more to pack, leave
            break
        # Everything is good, so we going to pick the object
        niryo_robot.pick_from_pose(*obj_pose.to_list())

        # Packing
        place_pose_high = place_pose.copy_with_offsets(z_offset=0.05)
        niryo_robot.move_pose(*place_pose_high.to_list())
        niryo_robot.move_pose(*place_pose.copy_with_offsets(z_offset=0.01).to_list())
        # Opening gripper to pack
        niryo_robot.open_gripper()
        # Going down to pick package
        niryo_robot.move_pose(*place_pose.copy_with_offsets(z_offset=-0.011).to_list())
        niryo_robot.close_gripper()
        niryo_robot.move_pose(*place_pose.copy_with_offsets(z_offset=0.05).to_list())

        # Getting place position
        if shape_obj == ObjectShape.SQUARE:
            target_pose = pos_conditioning1.copy_with_offsets(
                y_offset=0.05 * catch_count_category_1 % 3,
            )
            catch_count_category_1 += 1
        else:
            target_pose = pos_conditioning2.copy_with_offsets(
                y_offset=0.05 * catch_count_category_2 % 3,
            )
            catch_count_category_2 += 1

        niryo_robot.place_from_pose(*target_pose.to_list())

    niryo_robot.move_pose(*observation_pose_wks_pick.to_list())


if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Changing tool
    robot.update_tool()
    # Calibrate robot if robot needs calibration
    robot.calibrate_auto()
    # Launching main process
    process(robot)
    # Ending
    robot.go_to_sleep()
    # Releasing connection
    robot.close_connection()



#!/usr/bin/env python
# coding=utf-8


class PoseObject:
    """
    Pose object which stores x, y, z, roll, pitch & yaw parameters
    """

    def __init__(self, x, y, z, roll, pitch, yaw):
        # X (meter)
        self.x = float(x)
        # Y (meter)
        self.y = float(y)
        # Z (meter)
        self.z = float(z)
        # Roll (radian)
        self.roll = float(roll)
        # Pitch (radian)
        self.pitch = float(pitch)
        # Yaw (radian)
        self.yaw = float(yaw)

    def __str__(self):
        position = "x = {:.4f}, y = {:.4f}, z = {:.4f}".format(self.x, self.y, self.z)
        orientation = "roll = {:.3f}, pitch = {:.3f}, yaw = {:.3f}".format(self.roll, self.pitch, self.yaw)
        return position + "\n" + orientation

    def __repr__(self):
        return self.__str__()

    def copy_with_offsets(self, x_offset=0., y_offset=0., z_offset=0., roll_offset=0., pitch_offset=0., yaw_offset=0.):
        """
        Create a new pose from copying from copying actual pose with offsets

        :rtype: PoseObject
        """
        return PoseObject(self.x + x_offset,
                          self.y + y_offset,
                          self.z + z_offset,
                          self.roll + roll_offset,
                          self.pitch + pitch_offset,
                          self.yaw + yaw_offset)

    def to_list(self):
        """
        Return a list [x, y, z, roll, pitch, yaw] corresponding to the pose's parameters

        :rtype: list[float]
        """
        list_pos = [self.x, self.y, self.z, self.roll, self.pitch, self.yaw]
        return list(map(float, list_pos))


class HardwareStatusObject:
    """
    Object used to store every hardware information
    """

    def __init__(self, rpi_temperature, hardware_version, connection_up,
                 error_message, calibration_needed, calibration_in_progress,
                 motor_names, motor_types,
                 motors_temperature, motors_voltage, hardware_errors):
        # Number representing the rpi temperature
        self.rpi_temperature = rpi_temperature
        # Number representing the hardware version
        self.hardware_version = hardware_version
        # Boolean indicating if the connection with the robot is up
        self.connection_up = connection_up
        # Error message status on error
        self.error_message = error_message
        # Boolean indicating if a calibration is needed
        self.calibration_needed = calibration_needed
        # Boolean indicating if calibration is in progress
        self.calibration_in_progress = calibration_in_progress

        # Following list describe each motor
        # Row 0 for first motor, row 1 for second motor, row 2 for third motor, row 3 for fourth motor
        # List of motor names
        self.motor_names = motor_names
        # List of motor types
        self.motor_types = motor_types
        # List of motors_temperature
        self.motors_temperature = motors_temperature
        # List of motors_voltage
        self.motors_voltage = motors_voltage
        # List of hardware errors
        self.hardware_errors = hardware_errors

    def __str__(self):
        list_string_ret = list()
        list_string_ret.append("Temp (°C) : {}".format(self.rpi_temperature))
        list_string_ret.append("Hardware version : {}".format(self.hardware_version))
        list_string_ret.append("Connection Up : {}".format(self.connection_up))
        list_string_ret.append("Error Message : {}".format(self.error_message))
        list_string_ret.append("Calibration Needed : {}".format(self.calibration_needed))
        list_string_ret.append("Calibration in progress : {}".format(self.calibration_in_progress))
        list_string_ret.append("MOTORS INFOS : Motor1, Motor2, Motor3, Motor4, Motor5, Motor6,")
        list_string_ret.append("Names : {}".format(self.motor_names))
        list_string_ret.append("Types : {}".format(self.motor_types))
        list_string_ret.append("Temperatures : {}".format(self.motors_temperature))
        list_string_ret.append("Voltages : {}".format(self.motors_voltage))
        list_string_ret.append("Hardware errors : {}".format(self.hardware_errors))
        return "\n".join(list_string_ret)

    def __repr__(self):
        return self.__str__()


class DigitalPinObject:
    """
    Object used to store information on digital pins
    """

    def __init__(self, pin_id, name, mode, state):
        # Pin ID
        self.pin_id = pin_id
        # Name
        self.name = name
        # Input or output
        self.mode = mode
        # High or Low
        self.state = state

    def __str__(self):
        string_ret = "Pin : {}".format(self.pin_id)
        string_ret += ", Name : {}".format(self.name)
        string_ret += ", Mode : {}".format(self.mode)
        string_ret += ", State : {}".format(self.state)
        return string_ret

    def __repr__(self):
        return self.__str__()


class AnalogPinObject:
    """
    Object used to store information on digital pins
    """

    def __init__(self, pin_id, name, mode, value):
        # Pin ID
        self.pin_id = pin_id
        # Name
        self.name = name
        # Input or output
        self.mode = mode
        # Tension
        self.value = value

    def __str__(self):
        string_ret = "Pin : {}".format(self.pin_id)
        string_ret += ", Name : {}".format(self.name)
        string_ret += ", Mode : {}".format(self.mode)
        string_ret += ", State : {}".format(self.value)
        return string_ret

    def __repr__(self):
        return self.__str__()



"""
File containing functions for communication with Socket !
"""

import sys
import ast
import json
import struct

from .enums_communication import READ_SIZE, DEFAULT_PACKET_SIZE_INFOS


# --- RECEPTION -- #
def receive_data(sckt, packet_size_infos, buffer_size):
    """
    Receive msg which cannot be contained in only one buffer

    :param sckt: the socket
    :param packet_size_infos: Format de l'objet du message : permet de savoir sur cb de bytes est codee la taille
    :param buffer_size: buffer size for reading socket's buffer
    :return: Bytes corresponding to received packet !
    """
    nbr_bytes = packet_size_infos["nbr_bytes"]
    received_data = sckt.recv(nbr_bytes)
    if len(received_data) != nbr_bytes:
        return None
    size_packet = int(struct.unpack(packet_size_infos["type"], received_data)[0])
    received_big_data = "" if sys.version_info[0] == 2 else b""

    while len(received_big_data) < size_packet:
        received_big_data += sckt.recv(min(buffer_size, size_packet - len(received_big_data)))

    return received_big_data if sys.version_info[0] == 2 else received_big_data.decode()


def receive_dict(sckt, packet_size_infos=DEFAULT_PACKET_SIZE_INFOS, buffer_size=READ_SIZE):
    """
    Receive json through a Socket

    :param sckt: the socket
    :param packet_size_infos: Format de l'objet du message : permet de savoir sur cb de bytes est codee la taille
    :param buffer_size: buffer size for reading socket's buffer
    :return:
    """
    msg = receive_data(sckt, buffer_size=buffer_size, packet_size_infos=packet_size_infos)
    return data_to_dict(msg)


def receive_dict_w_payload(sckt, packet_size_infos=DEFAULT_PACKET_SIZE_INFOS, buffer_size=READ_SIZE):
    """
    Receive json through a Socket then extract payload content.
    Payload can typically be an heavy file like an image

    :param sckt: the socket
    :param packet_size_infos: Format de l'objet du message : permet de savoir sur cb de bytes est codee la taille
    :param buffer_size: buffer size for reading socket's buffer
    :return: dict of packet's JSON, payload
    """
    dict_data = receive_dict(sckt, buffer_size=buffer_size, packet_size_infos=packet_size_infos)
    payload_size = dict_data["payload_size"]
    received_payload = "" if sys.version_info[0] == 2 else b""
    while len(received_payload) < payload_size:
        received_payload += sckt.recv(buffer_size)

    return dict_data, received_payload


# - JSON - #
def data_to_dict(data):
    """
    Convert a string representing a JSON to a JSON

    :param data: String representing a JSON
    :return: dict of this JSON
    """
    if data is None:
        return None
    # if type(data) == bytes:
    #     data = data.decode()
    return ast.literal_eval(data)


def dict_to_packet(dict_obj, packet_size_infos=DEFAULT_PACKET_SIZE_INFOS):
    """
    Convert dict to packet

    :type dict_obj: dict
    :param packet_size_infos:
    :return: packet
    """
    json_obj = json.dumps(dict_obj)

    if sys.version_info[0] == 3:
        json_obj = json_obj.encode()
    packet = "" if sys.version_info[0] == 2 else b""
    packet += struct.pack(packet_size_infos["type"], len(json_obj))
    packet += json_obj
    return packet
